#!/usr/bin/Rscript
#
# dcVar.R - Bill White - 3/9/15
#
# An R script for comparing to dcVar implemented in C++ in inbix.

library(snpStats)

rm(list=ls())

# -----------------------------------------------------------------------------
#' Differential coexpression genetic association network algorithm
#' 
#' \code{dcgain} 
#' 
#' @param inbixData is a data frame with samples in rows, genes in columns
#' and phenotype in the last column
#' @return results a matrix of gene by gene differential coexpression values
#' @export
dcgain <- function(inbixData) {
  phenos <- inbixData[, ncol(inbixData)] + 1
  exprBySubj <- inbixData[, 1:(ncol(inbixData)-1)]
  exprByGene <- t(exprBySubj)
  varNames <- colnames(inbixData)[1:ncol(exprBySubj)]
  n1 <- length(which(phenos == 1))
  n2 <- length(which(phenos == 2))
  print(dim(inbixData))
  cat("Group 1:", n1, "Group 2:", n2, "\n")
  nVars <- ncol(exprBySubj)
  
  # determine group correlations
  expr_g1 <- exprByGene[, phenos == 1]
  expr_g2 <- exprByGene[, phenos == 2]
  #   cat("DEBUG Rinbix package\n")
  #   print(expr_g1[1:5, 1:5])
  #   print(expr_g2[1:5, 1:5])
  cor_table_g1 <- cor(t(expr_g1))
  cor_table_g2 <- cor(t(expr_g2))
  
  # main effect diagonal
  results <- matrix(nrow=nVars, ncol=nVars)
  pvalues <- matrix(nrow=nVars, ncol=nVars)
  for(i in 1:nrow(exprByGene)) {
    g1_data <- exprByGene[i, phenos == 1]
    g2_data <- exprByGene[i, phenos == 2]
    g1_mean <- mean(g1_data)
    g2_mean <- mean(g2_data)
    # z-test
    z_i_1 <- 0.5 * log((abs((1 + g1_mean) / (1 - g1_mean))))
    z_i_2 <- 0.5 * log((abs((1 + g2_mean) / (1 - g2_mean))))
    Z_i <- abs(z_i_1 - z_i_2) / sqrt((1/(n1 - 3) + 1 / (n2 - 3)))
    #Z_i <- abs(g1_mean - g2_mean) / sqrt((1/(n1 - 3) + 1 / (n2 - 3)))
    results[i, i] <- Z_i
    pvalues[i, i] <- 1
    # t-test
    #     t_result <- t.test(g1_data, g2_data)
    #     t_i <- abs(t_result$statistic)
    #     results[i, i] <- t_i
    #     pvalues[i, i] <- t_result$p.value
    
    #cat(Z_i, t_i, "\n")
  }
  
  # ----------------------------------------------------------------------------
  # interactions
  for(i in 1:nVars) {
    for(j in 1:nVars) {
      if(j <= i) {
        next
      }
      r_ij_1 <- cor_table_g1[i, j]
      r_ij_2 <- cor_table_g2[i, j]
      z_ij_1 <- 0.5 * log((abs((1 + r_ij_1) / (1 - r_ij_1))))
      z_ij_2 <- 0.5 * log((abs((1 + r_ij_2) / (1 - r_ij_2))))
      Z_ij <- abs(z_ij_1 - z_ij_2) / sqrt((1/(n1 - 3) + 1 / (n2 - 3)))
      pval <- 2 * pnorm(-abs(Z_ij))
      results[i, j] <- Z_ij
      results[j, i] <- Z_ij
      pvalues[i, j] <- pval
      pvalues[j, i] <- pval
    }
  }
  colnames(results) <- varNames
  colnames(pvalues) <- varNames
  
  list(scores=results, pvals=pvalues)
}

# -----------------------------------------------------------------------------
# read data
# variants
#variantsObj <- read.plink("rs235214.bed")
variantsObj <- read.plink("variants10.bed")
#variantsObj <- read.plink("variants100.bed")
#variantsObj <- read.plink("variants.bed")
subjIds <- rownames(variantsObj$fam)
snps <- variantsObj$genotypes
snpNames <- colnames(snps)
numVariants <- ncol(snps)
  
# gene expression
expression_num <- read.table("expression100.matrix", sep="\t", head=T)
#expression_num <- read.table("expression.matrix", sep="\t", head=T)
expression_num_bed_order <- expression_num[match(subjIds, expression_num[,2]), ]
expr <- expression_num_bed_order[, 3:ncol(expression_num)]
geneNames <- colnames(expr)
  
# -----------------------------------------------------------------------------
# prepare for looping over variants and expression pairs
fdrValue = 0.05
nGenes <- ncol(expr)
nCombs = choose(nGenes, 2)
correctedP = fdrValue / (nCombs * numVariants);
cat("FDR Corrected p-value:", correctedP, "\n")

# -----------------------------------------------------------------------------
# build a new phenotype from variant genotypes
mapPhenosToModel <- function(phenos, varModel) {
  if(varModel == "dom") {
    mappedPhenos <- ifelse(phenos == 3, 1, 0)
  } else {
    if(varModel == "rec") {
      mappedPhenos <- ifelse(phenos == 1, 1, 0)
    } else {
      # hom
      mappedPhenos <- ifelse(phenos == 2, 1, -9)
    }
  }
  mappedPhenos
}

# -----------------------------------------------------------------------------
# loop over variants and expression pairs
# dom, rec, hom
variantModel <- "dom"
for(variantIdx in 1:numVariants) {
  variantName <- snpNames[variantIdx]
  
  # filter missing phenotypes
  goodSubjIdx <- which(variantsObj$fam$affected != -9)
  goodExpr <- expr[goodSubjIdx, ]
    
  # create phenotype from SNPs and variant model
  variantSnps <- snps[, variantName]
  goodVariants <- variantSnps[goodSubjIdx]
  mappedPhenos <- mapPhenosToModel(as.integer(goodVariants), variantModel)
  
  # remove subjects with missing mapped phenotypes
  goodSubjIdx <- which(mappedPhenos != -9)
  cat("Read ", length(goodSubjIdx), "non missing phenotypes\n")
  goodExpr <- goodExpr[goodSubjIdx, ]
  
  # run dcGAIN
  classificationData <- cbind(goodExpr, mappedPhenos)
  res <- dcgain(classificationData)
  
  # pull out expression pairs that pass FDR corrected p-values
  finalRes <- NULL
  goodCount <- 0
  for(i in 1:nGenes) {
    for(j in 1:nGenes) {
      if(j <= i) { 
        next
      }
      thisP <- res$pvals[i, j]
      if(thisP < correctedP) {
        finalRes <- rbind(finalRes, data.frame(gene1=geneNames[i], 
                                               gene2=geneNames[j], 
                                               pval=thisP))
        goodCount <- goodCount + 1
      }
    }
  }
  cat("Found [", goodCount, "] FDR tested p-values, min/max", 
      min(as.vector(res$pvals)), "/", max(as.vector(res$pvals)), "\n")
  
  # save results by variant to file
  thisFilename <- paste(variantName, ".dcVarRTest.txt", sep="")
  write.table(finalRes, thisFilename, sep="\t", quote=F, col.names=F, row.names=F)
}
